#include "smear.h"


static	void
arrayCheckAllocation(ArrayStructure *asp, int offset)
{
	int	allocCount;
	double	*dp;

	if (offset < 0 ||
	    offset >= MAX_ARRAY_OFFSET ||
	    offset < asp->asAllocated)
		return;

	allocCount = asp->asAllocated;

	while (allocCount <= offset)
		allocCount += ARRAY_ALLOCATION_CHUNK;

	dp = (double *)malloc(allocCount * sizeof(double));

	if (asp->asAllocated)
	{
		memcpy(dp, asp->asData, sizeof(double) * asp->asAllocated);
		free(asp->asData);
	}

	memset(((char *)dp) + asp->asAllocated * sizeof(double), '\0',
	       sizeof(double)(allocCount - asp->asAllocated));

	asp->asAllocated = allocCount;
	asp->asData = dp;
}


void
arrayFreeAllocation(ArrayStructure *asp)
{
	if (asp->asAllocated && asp->asData)
	{
		asp->asAllocated = 0;
		free(asp->asData);
		asp->asData = (double *)0;
	}
}


ArrayStructure	*
arrayAllocateNew(void)
{
	ArrayStructure	*asp;

	if ((asp = (ArrayStructure *)malloc(sizeof(ArrayStructure)))
	    == (ArrayStructure *)0)
		return asp;

	asp->asAllocated = 0;

	return asp;
}


ArrayAccess	*
arrayAllocateNewAccess(void)
{
	return (ArrayAccess *)malloc(sizeof(ArrayAccess));
}


double	*
arrayGetEvaluationPointer(ArrayStructure *asp, double offset)
{
static	double	errorReturn;
	int	ioff;

	ioff = (int)floor(offset);

	if (ioff < 0 || ioff >= MAX_ARRAY_OFFSET)
	{
		errorReturn = 0.0;
		return &errorReturn;
	}

	arrayCheckAllocation(asp, ioff);

	return asp->asData + ioff;
}


double
arrayEvaluate(ArrayStructure *asp, double offset)
{
	return *(arrayGetEvaluationPointer(asp, offset));
}
