#include "smear.h"

__inline	int
BasicRandom(int n)
{
	return rand() % n;
}


__inline	int
WeightedRandom(int n)
{
	int n2 = BasicRandom(n * n);
	int w;

	w = n - (int)sqrt(n2) - 1;

	if (w < 0)
		w = 0;

	if (w >= n)
		w = n - 1;

	return w;
}


void
LoadAllScripts(char *dir, char *match, Environment *list[], int *count)
{
	WIN32_FIND_DATA		fnd;
	HANDLE				fnd_handle;
	DWORD				taa;				
	Environment			*ep;

	chdir(dir);
	fnd_handle = FindFirstFile(match, &fnd);

	*count = 0;

	if (fnd_handle != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (((taa = GetFileAttributes(fnd.cFileName)) == 0xFFFFFFFF) || (taa & FILE_ATTRIBUTE_DIRECTORY))
				continue;

			if ((ep = ParseFile(fnd.cFileName)) == (Environment *)0)
				continue;

			ep->eFile = strdup(fnd.cFileName);

			list[(*count)++] = ep;

		} while (FindNextFile(fnd_handle, &fnd) == TRUE);

		FindClose(fnd_handle);
	}

	chdir("..");

	return;
}

Environment	*moveEnvironments[MAX_ENVIRONMENTS];
int			moveEnvironmentCount = -1;

void
LoadAllMoveScripts(void)
{
	Environment	*ep;

	LoadAllScripts("Moves", "*.sms", moveEnvironments, &moveEnvironmentCount);

	if (moveEnvironmentCount == 0)
	{
		ParserSetupDummyScript("(declare init (set $usePolar 1) (set $useAspect 1) (set $a0 (+ 2 (rnd 16))) (set $a1 (+ .002 (rnd .02))) (set $a2 (+ .002 (rnd .006)))) (declare move (set $srcRadius (pos (- $radius $a2))) (set $srcTheta (+ $theta (* $a1 (sin (* $a0 $radius))))))");

		if ((ep = ParseFile((char *)0)) == (Environment *)0)
			return;

		moveEnvironments[moveEnvironmentCount++] = ep;
	}
}


Environment	*
ChooseMoveScript(void)
{
	int		w;
	Environment	*p;
static	Environment	*specialEnvironment = (Environment *)0;
	char		specialName[256];

	if (specialEnvironment != (Environment *)0)
	{
		// We should free everything up here, but, well.. I'm not going to
		// free(specialEnvironment);
		specialEnvironment =  (Environment *)0;
	}

	if (ssmGetNextMove(specialName))
		if ((specialEnvironment = ParseFile(specialName)) != (Environment *)0)
			return specialEnvironment;

	if (moveEnvironmentCount == 0)
		return (Environment *)0;

	w = WeightedRandom(moveEnvironmentCount);

	p = moveEnvironments[w];

	if (w < moveEnvironmentCount - 1)
	{
		int	loop;

		for (loop = w; loop < moveEnvironmentCount - 1; loop++)
			moveEnvironments[loop] = moveEnvironments[loop + 1];

		moveEnvironments[moveEnvironmentCount - 1] = p;
	}

	return p;
}




Environment	*colorEnvironments[MAX_ENVIRONMENTS];
int			colorEnvironmentCount = -1;

void
LoadAllColorScripts(void)
{
	Environment	*ep;

	LoadAllScripts("Color", "*.scs", colorEnvironments, &colorEnvironmentCount);

	if (colorEnvironmentCount == 0)
	{
		ParserSetupDummyScript("(declare init (set $useHSV 1) (set $useTime 1)) (declare color (set $h (* $time 0.05)) (set $v (* $value 2.0)) (set $s (* 2 (- 1 $value))))");

		if ((ep = ParseFile((char *)0)) == (Environment *)0)
			return;

		colorEnvironments[colorEnvironmentCount++] = ep;
	}
}


Environment	*
ChooseColorScript(void)
{
	int		w;
	Environment	*p;
static	Environment	*specialEnvironment = (Environment *)0;
	char		specialName[256];

	if (specialEnvironment != (Environment *)0)
	{
		// We should free everything up here, but, well.. I'm not going to
		// free(specialEnvironment);
		specialEnvironment =  (Environment *)0;
	}

	if (ssmGetNextColor(specialName))
		if ((specialEnvironment = ParseFile(specialName)) != (Environment *)0)
			return specialEnvironment;


	if (colorEnvironmentCount == 0)
		return (Environment *)0;

	w = WeightedRandom(colorEnvironmentCount);

	p = colorEnvironments[w];

	if (w < colorEnvironmentCount - 1)
	{
		int	loop;

		for (loop = w; loop < colorEnvironmentCount - 1; loop++)
			colorEnvironments[loop] = colorEnvironments[loop + 1];

		colorEnvironments[colorEnvironmentCount - 1] = p;
	}

	return p;
}


Environment	*waveformEnvironments[MAX_ENVIRONMENTS];
int			waveformEnvironmentCount = -1;

void
LoadAllWaveFormScripts(void)
{
	Environment	*ep;

	LoadAllScripts("WaveForm", "*.sws", waveformEnvironments, &waveformEnvironmentCount);

	if (waveformEnvironmentCount == 0)
	{
		ParserSetupDummyScript("(declare init (set $paths 1) (set $steps 256)) (declare newline (set $angle (+ $angle 0.015))) (declare waveform (polar $radius $theta (- (* $step 2.0) 1.0) $mid) (cart $x0 $y0 $radius (+ $theta $angle)))");

		if ((ep = ParseFile((char *)0)) == (Environment *)0)
			return;

		waveformEnvironments[waveformEnvironmentCount++] = ep;
	}
}


Environment	*
ChooseWaveFormScript(void)
{
	int		w;
	Environment	*p;
static	Environment	*specialEnvironment = (Environment *)0;
	char		specialName[256];

	if (specialEnvironment != (Environment *)0)
	{
		// We should free everything up here, but, well.. I'm not going to
		// free(specialEnvironment);
		specialEnvironment =  (Environment *)0;
	}

	if (ssmGetNextWaveform(specialName))
		if ((specialEnvironment = ParseFile(specialName)) != (Environment *)0)
			return specialEnvironment;


	if (waveformEnvironmentCount == 0)
		return (Environment *)0;

	w = WeightedRandom(waveformEnvironmentCount);

	p = waveformEnvironments[w];

	if (w < waveformEnvironmentCount - 1)
	{
		int	loop;

		for (loop = w; loop < waveformEnvironmentCount - 1; loop++)
			waveformEnvironments[loop] = waveformEnvironments[loop + 1];

		waveformEnvironments[waveformEnvironmentCount - 1] = p;
	}

	return p;
}


Environment	*particleEnvironments[MAX_ENVIRONMENTS];
int			particleEnvironmentCount = -1;

void
LoadAllParticleScripts(void)
{
	LoadAllScripts("Particle", "*.sps", particleEnvironments, &particleEnvironmentCount);
}


Environment	*
ChooseParticleScript(void)
{
	int		w;
	Environment	*p;
static	Environment	*specialEnvironment = (Environment *)0;
	char		specialName[256];

	if (specialEnvironment != (Environment *)0)
	{
		// We should free everything up here, but, well.. I'm not going to
		// free(specialEnvironment);
		specialEnvironment =  (Environment *)0;
	}

	if (ssmGetNextParticle(specialName))
		if ((specialEnvironment = ParseFile(specialName)) != (Environment *)0)
			return specialEnvironment;

	if (particleEnvironmentCount == 0)
		return (Environment *)0;

	w = WeightedRandom(particleEnvironmentCount);

	p = particleEnvironments[w];

	if (w < particleEnvironmentCount - 1)
	{
		int	loop;

		for (loop = w; loop < particleEnvironmentCount - 1; loop++)
			particleEnvironments[loop] = particleEnvironments[loop + 1];

		particleEnvironments[particleEnvironmentCount - 1] = p;
	}

	return p;
}


void
LoadEverything(void)
{
	LoadAllMoveScripts();
	LoadAllColorScripts();
	LoadAllWaveFormScripts();
	LoadAllParticleScripts();
}
