//////////////////////////////////////////////////////////////////////////////////////
//
// Sonique Visual Plugin Wrapper File
//
//   (c) 1999 Sonique Team
//       Edited by Rajest Singh (shock@ihug.com.au)
//       Edited futher by mykel (mykel@earthling.net)
//
//////////////////////////////////////////////////////////////////////////////////////

#include "smear.h"

#define	MAP_MERGING
#define	MAP_SUPERMERGE

#if defined(MAP_SUPERMERGE) && !defined(MAP_MERGING)
# define	MAP_MERGING
#endif

#define	SIMPLE_RAND(n) (rand() % n)

Environment	*_global_move_ep;

unsigned long	colorMap[256];
unsigned long	fadeMaps[2][256];


//////////////////////////////////////////////////////////////////////////////////////
//
// Vis Globals
//
//////////////////////////////////////////////////////////////////////////////////////
unsigned long	*_Video;
int		_width;
int		_height;
int		_myHeight;
int		_pitch;
int		_div;
VisData		*_pVD;

int		_Ymin;
int		_Ymax;



//////////////////////////////////////////////////////////////////////////////////////
//
// Vis Argument Variables
//
//////////////////////////////////////////////////////////////////////////////////////
static	int	ARG_SampleInt = 0;
static	int	ARG_SampleTF = 0;

static	int	ARG_MovemapHoldTime = 0;
static	int	ARG_ColormapHoldTime = 0;
static	int	ARG_ColormapFadeTime = 0;
static	int	ARG_WaveformHoldTime = 0;
static	int	ARG_WaveformFadeTime = 0;
static	int	ARG_ParticleHoldTime = 0;
static	int	ARG_ParticleFadeTime = 0;
static	int	ARG_ParticleChance = 0;
static	int	ARG_LetterBoxPercent = 0;
static	int	ARG_ShowTitles = 0;
#ifdef	USE_MAP_THREADS
static	int	ARG_ThreadMaps = 0;
#endif
#ifdef	MAP_MERGING
static	int	ARG_MovemapFadeTime = 0;
#endif
#ifdef	MAP_SUPERMERGE
static	int	ARG_MovemapConstantFade = 0;
static	int	ARG_TotalMayhem = 0;
#endif
	int	ARG_DebugScripts = 0;
static	int	ARG_NoVisName = 0;

static	int	WRITE_SampleInt = 0;
static	int	WRITE_SampleTF = 0;

static	int	WRITE_MovemapHoldTime = 0;
static	int	WRITE_ColormapHoldTime = 0;
static	int	WRITE_ColormapFadeTime = 0;
static	int	WRITE_WaveformHoldTime = 0;
static	int	WRITE_WaveformFadeTime = 0;
static	int	WRITE_ParticleHoldTime = 0;
static	int	WRITE_ParticleFadeTime = 0;
static	int	WRITE_ParticleChance = 0;
static	int	WRITE_LetterBoxPercent = 0;
static	int	WRITE_ShowTitles = 0;
#ifdef	USE_MAP_THREADS
static	int	WRITE_ThreadMaps = 0;
#endif
#ifdef	MAP_MERGING
static	int	WRITE_MovemapFadeTime = 0;
#endif
#ifdef	MAP_SUPERMERGE
static	int	WRITE_MovemapConstantFade = 0;
static	int	WRITE_TotalMayhem = 0;
#endif
static	int	WRITE_DebugScripts = 0;
static	int	WRITE_NoVisName = 0;

static	double	VALUE_ParticleChance;


//////////////////////////////////////////////////////////////////////////////////////
//
// DwordCopy_Inline(unsigned long *pDst, unsigned long *pSrc, unsigned long numDwords)
//
// Fast copy numDwords double words from pSrc to pDst.  Considerably faster than memcpy
// and friends.
//
//////////////////////////////////////////////////////////////////////////////////////
__inline void
DwordCopy_Inline(unsigned long* pDst, unsigned long* pSrc, unsigned long numDwords)
{
	__asm
	{
		push EBP
		mov EDI, pDst
		mov ESI, pSrc
		mov EBP, numDwords
		push EBP
		shr EBP, 2
		jz WordCopy
		sub EDI, 16
	 Top:
		add EDI, 16
		mov EAX, [ESI+0]
		mov EBX, [ESI+4]
		mov ECX, [ESI+8]
		mov EDX, [ESI+12]
		add ESI, 16
		dec EBP
		mov [EDI+0], EAX
		mov [EDI+4], EBX
		mov [EDI+8], ECX
		mov [EDI+12], EDX
		jnz Top
		add EDI, 16
	 WordCopy:
		pop EBP
 		sub EDI, 4
		and EBP, 3
		jz NoSmall
	 TopSmall:
		add EDI, 4
		mov EAX, [ESI]
		add ESI, 4
		dec EBP
		mov [EDI], EAX
		jnz TopSmall
	 NoSmall:
		pop EBP
	};
}



//////////////////////////////////////////////////////////////////////////////////////
//
// map gloabls
//
//////////////////////////////////////////////////////////////////////////////////////

// Two of each.  One current, one being built.
#ifdef	MAP_SUPERMERGE
unsigned long	*mapsData[3] = { (unsigned long *)0, (unsigned long *)0,
				 (unsigned long *)0 };	// Current and new, next
#else
unsigned long	*mapsData[2] = { (unsigned long *)0, (unsigned long *)0 };	// Current and new
#endif
int				mapWidth = -1;
int				mapHeight = -1;

static char			*clickedMoveName = (char *)0;

mapFunctionExternalSetup()
{
	// Setup the move name here, before we start building a new one
	if (_global_move_ep)
		clickedMoveName = _global_move_ep->eFile;
	else
		clickedMoveName = "None";

	_global_move_ep = ChooseMoveScript();
	if (_global_move_ep)
		SetupForRun(_global_move_ep);
}


//////////////////////////////////////////////////////////////////////////////////////
//
// maps
//
// Handle map lists.  Various build functions.
//
//////////////////////////////////////////////////////////////////////////////////////

int				mapLine = 0;
int				mapStop = 0;
int				mapFinished = 0;
HANDLE				mapThread;
HANDLE				mapMutex;
int				mapRunning = 0;
int				mapEndTime = 0;
int				mapValid = 1;
int				mapStartIn = 0;
int				mapRestart = 0;
int				mapRebuildTime = MAP_SLOWBUILD;

#ifdef	MAP_MERGING
int				mapMerging = 0;
int				mapMergeTime;
int				mapMergeEnd;
int				mapMergeFactor;
#endif


// Fix the allocation, return non-zero if map data was reallocated
int
mapsCheckAllocation(void)
{
	int	i;

	if (mapWidth == _width &&
		mapHeight == _myHeight &&
		mapsData[0] != (unsigned long *)0 &&
		mapsData[1] != (unsigned long *)0)
		  return 0;

#ifdef	USE_MAP_THREADS
	if (ARG_ThreadMaps)
		parentStopCurrentBuild();
#endif

	mapWidth = _width;
	mapHeight = _myHeight;

#ifdef	MAP_SUPERMERGE
	for (i = 0; i < (ARG_MovemapConstantFade ? 3 : 2); i++)
#else
	for (i = 0; i < 2; i++)
#endif
	{
		if (mapsData[i] != (unsigned long *)0)
			free(mapsData[i]);

		mapsData[i] = (unsigned long *)malloc(mapWidth * mapHeight * sizeof(unsigned long) * 2);

		memset(mapsData[i], '\0', sizeof(unsigned long) * mapWidth * mapHeight * 2);
	}

	mapLine = 0;

	if (ARG_NoVisName)
	{
		unsigned	long	*mp;
		int			yadd;
		int			xadd;
		int			yv;
		int			x, y;

		mp = mapsData[0];

		yadd = mapHeight / 2;
		xadd = mapWidth / 2;
		// Build the quickest and dirtiest map here..
		for (y = 0; y < mapHeight; y++)
		{
			yv = (y - yadd) * 64500 + (yadd << 16);
			for (x = 0; x < mapWidth; x++)
			{
				*(mp++) = (x - xadd) * 64500 + (xadd << 16);
				*(mp++) = yv;
			}
		}
#ifdef	MAP_SUPERMERGE
		if (ARG_MovemapConstantFade)
			memcpy(mapsData[1], mapsData[0], mapWidth * mapHeight * sizeof(unsigned long) * 2);
#endif 
	}

	return 1;
}


// Swap the map pointers.
__inline void
mapsSwap(void)
{
	unsigned long	*temp;

#ifdef	MAP_SUPERMERGE
	if (ARG_MovemapConstantFade)
	{
		temp = mapsData[0];
		mapsData[0] = mapsData[1];
		mapsData[1] = mapsData[2];
		mapsData[2] = temp;

		return;
	}
#endif

	temp = mapsData[0];
	mapsData[0] = mapsData[1];
	mapsData[1] = temp;
}


#ifdef	USE_OLD_LINE_BASED_MAPS
// Build the next line in the new map.
//
// Return value:
//	0	Success
//	1	Success - Job Complete
//	-1	Failed
int
mapsBuildNextLine(void)
{
	if (mapLine >= mapHeight)
		return 1;


	if (_global_move_ep)
		RunForLine(_global_move_ep, mapsData[1], _width, _myHeight, mapLine);

	mapLine++;

	return 0;
}

#else

int
mapsTimedBuildLines(int ms)
{
	if (!_global_move_ep)
		return -1;

	do
	{
		if (mapLine >= mapHeight)
			return 1;

#ifdef	MAP_SUPERMERGE
		if (TimedRunForLine(_global_move_ep, mapsData[ARG_MovemapConstantFade ? 2 : 1], _width, _myHeight, mapLine, &ms) == 0)
			mapLine++;
#else
		if (TimedRunForLine(_global_move_ep, mapsData[1], _width, _myHeight, mapLine, &ms) == 0)
			mapLine++;
#endif

	} while (ms > 0);

	return 0;
}

#endif

// Handle the map stuff - just do it
void
mapsHandleMapsNonThread(void)
{
	int	now;
	int	rv;

	now = GetTickCount();

	if (ssmCheckNextMove())
	{
		// Build a complete map
		mapFunctionExternalSetup();
		mapLine = 0;
		mapRebuildTime = MAP_FASTBUILD;
		mapEndTime = now;
		mapMergeTime = 100 * ARG_MovemapFadeTime * (SIMPLE_RAND(100) + 50) / 100;
	}


	if (mapsCheckAllocation() == 1)
	{
		if (ARG_NoVisName)
		{
			mapMerging = 0;
			mapFunctionExternalSetup();
			mapRebuildTime = (MAP_SLOWBUILD + MAP_FASTBUILD) >> 1;
			mapLine = 0;
			mapEndTime = now + 100 * ARG_MovemapHoldTime * (SIMPLE_RAND(100) + 50) / 200;
			mapMergeTime = 100 * ARG_MovemapFadeTime * (SIMPLE_RAND(100) + 50) / 100;
		}
		else
		{
			// Build a complete map
			mapFunctionExternalSetup();
			mapLine = 0;
#ifdef	USE_OLD_LINE_BASED_MAPS
			while ((rv = mapsBuildNextLine()) == 0)
				;
#else
			mapValid = 0;
#endif
			mapRebuildTime = MAP_FASTBUILD;
			mapMerging = 0;

			mapEndTime = now + 1500;
			mapMergeTime = 0;
		}
	}

#ifdef	MAP_SUPERMERGE

	if (ARG_MovemapConstantFade)
	{
		mapMerging = 1;
		mapMergeFactor = (mapHeight - mapLine) * 256 / mapHeight;
		if (!ARG_TotalMayhem)
		{
			int	t = (mapEndTime - now) * 256 / mapMergeTime;

			if (t > mapMergeFactor)
				mapMergeFactor = t;
		}

		if (mapMergeFactor <= 0)
			mapMergeFactor = 1;
		if (mapMergeFactor > 255)
			mapMergeFactor = 255;

		if ((rv = mapsTimedBuildLines(mapRebuildTime)) == 1 &&
		    (now >= mapEndTime || ARG_TotalMayhem))
		{
			mapValid = 1;
			mapMerging = 1;
			mapMergeFactor = 255;

			mapsSwap();
			mapFunctionExternalSetup();
			if (ARG_TotalMayhem)
				mapRebuildTime = (MAP_FASTBUILD + MAP_SLOWBUILD) >> 1;
			else
				mapRebuildTime = MAP_SLOWBUILD;
			mapLine = 0;

			if (ARG_TotalMayhem)
				mapEndTime = now + 1;
			else
				mapEndTime = now + 100 * ARG_MovemapHoldTime * (SIMPLE_RAND(100) + 50) / 200;
			mapMergeTime = mapEndTime - now;
		}

		if (rv == -1)
		{
			_global_move_ep = ChooseMoveScript();
			if (_global_move_ep)
				SetupForRun(_global_move_ep);

			if (ARG_TotalMayhem)
				mapRebuildTime = (MAP_FASTBUILD + MAP_SLOWBUILD) >> 1;
			else
				mapRebuildTime = MAP_SLOWBUILD;

			mapLine = 0;
		}

		return;
	}
#endif



#ifdef	MAP_MERGING
	if (mapMerging)
	{
		if (now >= mapMergeEnd || mapMergeTime <= 0)
			mapMerging = 0;
		else
		{
			mapMergeFactor = (mapMergeEnd - now) * 256 / mapMergeTime;
			if (mapMergeFactor > 255)
				mapMergeFactor = 255;
			if (mapMergeFactor <= 0)
				mapMerging = 0;
		}

		if (mapMerging == 0)
		{
			mapsSwap();
			mapFunctionExternalSetup();
			mapRebuildTime = MAP_SLOWBUILD;
			mapLine = 0;
			mapEndTime = now + 100 * ARG_MovemapHoldTime * (SIMPLE_RAND(100) + 50) / 100;
			mapMergeTime = 100 * ARG_MovemapFadeTime * (SIMPLE_RAND(100) + 50) / 100;
		}
		else
			return;
	}
#endif

#ifdef	USE_OLD_LINE_BASED_MAPS
	if ((rv = mapsBuildNextLine(mapRebuildTime)) == 1)
#else
	if ((rv = mapsTimedBuildLines(mapRebuildTime)) == 1)
#endif
	{
		if (now >= mapEndTime)
		{
#ifdef	MAP_MERGING
			mapValid = 1;
			mapMerging = 1;
			if (mapMergeTime > 3 * ARG_MovemapFadeTime * 100 / 2)
				mapMergeTime = 3 * ARG_MovemapFadeTime * 100 / 2;
			mapMergeEnd = now + mapMergeTime;
			mapMergeFactor = 255;
#else
			mapValid = 1;
			mapsSwap();
			mapFunctionExternalSetup();
			mapRebuildTime = MAP_SLOWBUILD;
			mapLine = 0;
			mapEndTime = now + 100 * ARG_MovemapHoldTime * (SIMPLE_RAND(100) + 50) / 100;
#endif
		}
	}

	if (rv == -1)
	{
		_global_move_ep = ChooseMoveScript();
		if (_global_move_ep)
			SetupForRun(_global_move_ep);

		mapRebuildTime = MAP_SLOWBUILD;
		mapLine = 0;
		mapEndTime = now + 100 * ARG_MovemapHoldTime * (SIMPLE_RAND(100) + 50) / 100;
		mapMergeTime = 100 * ARG_MovemapFadeTime * (SIMPLE_RAND(100) + 50) / 100;
	}
}



#ifdef	USE_MAP_THREADS
void
mapsHandleMapsThread(void)
{
//	int	rv;
	int	now;

	now = GetTickCount();

	if (mapsCheckAllocation() == 1)
	{
		// The Check has already (hopefully) stopped the thread
		// All we need to do is start a new thread build going
		// at the increased speed
		parentStartNewBuild(1);
		mapEndTime = 0;
		mapValid = 0;
		mapRestart = 0;
	}

	if (mapRestart)
	{
		mapRestart = 0;
		parentStartNewBuild(0);
	}

	if ((now >= mapEndTime || !mapValid) && !parentIsBuilding())
	{
		mapsSwap();
		mapValid = 1;
		mapEndTime = now + 100 * ARG_MovemapHoldTime * (SIMPLE_RAND(100) + 50) / 100;
		parentStartNewBuild(0);
	}
}
#endif


void
mapsHandleMaps(void)
{
#ifdef	USE_MAP_THREADS
	if (ARG_ThreadMaps)
		mapsHandleMapsThread();
	else
#endif
		mapsHandleMapsNonThread();
}


//////////////////////////////////////////////////////////////////////////////////////
//
// sb (screen buffer)
//
// Functions for handling the screen buffer.  This is a 0-255 value buffer that maps
// directly to the screen.
//
//////////////////////////////////////////////////////////////////////////////////////
unsigned char	*sbData[2] = { (unsigned char *)0, (unsigned char *)0 };
unsigned char	*sbAllocs[2] = { (unsigned char *)0, (unsigned char *)0 };
int				sbWidth = -1;
int				sbHeight = -1;

unsigned char	sbFadeValue[256];

void
sbInitFadeValue(void)
{
	int	i;

	for (i = 0; i < 256; i++)
		sbFadeValue[i] = (int)((double)i * (0.95 + (double)SIMPLE_RAND(500) / 10000.0));
}


int
sbCheckAllocation(void)
{
	int	i;

	if (sbWidth == -1)
		sbInitFadeValue();

	if (sbWidth == _width &&
		sbHeight == _myHeight &&
		sbData[0] != (unsigned char *)0 &&
		sbData[1] != (unsigned char *)0)
		  return 0;

	sbWidth = _width;
	sbHeight = _myHeight;

	for (i = 0; i < 2; i++)
	{
		if (sbAllocs[i] != (unsigned char *)0)
			free(sbAllocs[i]);

		sbAllocs[i] = (unsigned char *)malloc(sbWidth * sbHeight * sizeof(unsigned char) + 2 * sbWidth + 2);

		memset(sbAllocs[i], '\0', sbWidth * sbHeight * sizeof(unsigned char) + 2 * sbWidth + 2);
		sbData[i] = sbAllocs[i] + sbWidth + 1;
	}

	return 1;
}


// Swap the sb pointers.
__inline void
sbSwap(void)
{
	unsigned char	*temp;

	temp = sbData[0];
	sbData[0] = sbData[1];
	sbData[1] = temp;
}

void
OLDsbApplyMove(void)
{
	unsigned long	*mp;
	unsigned long	*mp0;
	unsigned long	*mp1;
	unsigned char	*op;
	int				i;
	register int	x;
	register int	y;
	unsigned char			*offsetData;
	int				v1;
	int				v2;

	mapsHandleMaps();

	sbCheckAllocation();

	if (mapValid == 0)
	{
		int	size;

		memset(sbData[0], '\0', sbWidth * sbHeight);

		if (WSInit(_width, _height, WS_TITLE_LAYER))
		{
			size = _myHeight / 8;
			WSSetFont(colorMap[255], "verdana", size * 15 / 10, FW_BOLD, WS_TITLE_LAYER);
			WSSupDrawText(_width / 2, _myHeight / 2 - 2 * size, "Smear", WS_TITLE_LAYER, TA_CENTER);
			WSSetFont(colorMap[255], "verdana", size * 9 / 10, FW_BOLD, WS_TITLE_LAYER);
			WSSupDrawText(_width / 2, _myHeight / 2, "mykel :: xplo", WS_TITLE_LAYER, TA_CENTER);

			WSCopyOut(WS_TITLE_LAYER);
		}

		WSCopyHoldToSB(sbData[0], 255, WS_TITLE_LAYER);

		sbSwap();

		return;
	}

#define	INTERPOLATE16(a, b, v)	((a) + ((((b) - (a)) * (v)) >> 16))

#ifdef	MAP_MERGING
	if (mapMerging && mapMergeFactor > 0 && mapMergeFactor < 255)
	{
		mp0 = mapsData[0];
		mp1 = mapsData[1];
		op = sbData[1];

		for (i = sbWidth * sbHeight; i > 0; i--, op++, mp0++, mp1++)
		{
			x = (((*mp1++) >> 4) * (256 - mapMergeFactor) + ((*mp0++) >> 4) * mapMergeFactor) >> 4;
			y = (((*mp1) >> 4) * (256 - mapMergeFactor) + ((*mp0) >> 4) * mapMergeFactor) >> 4;

			// offsetData = pxlXYPixelPointer(sbData[0], (x >> 16), (y >> 16));
			//  um.. I think this is exactly the same, but hey..
			offsetData = sbData[0] + (x >> 16) + *(yCoordinates + (y >> 16));

			v1 = INTERPOLATE16(*(offsetData), *(offsetData + 1), x & 0xffff);
			v2 = INTERPOLATE16(*(offsetData + sbWidth), *(offsetData + sbWidth + 1), x & 0xffff);

			*op = sbFadeValue[INTERPOLATE16(v1, v2, y & 0xffff)];
		}
	}
	else
	{
#endif
		mp = mapsData[0];
		op = sbData[1];

		for (i = sbWidth * sbHeight; i > 0; i--, op++, mp++)
		{
			x = *mp++;
			y = *mp;

			// offsetData = pxlXYPixelPointer(sbData[0], (x >> 16), (y >> 16));
			//  um.. I think this is exactly the same, but hey..
			offsetData = sbData[0] + (x >> 16) + *(yCoordinates + (y >> 16));

			v1 = INTERPOLATE16(*(offsetData), *(offsetData + 1), x & 0xffff);
			v2 = INTERPOLATE16(*(offsetData + sbWidth), *(offsetData + sbWidth + 1), x & 0xffff);

			*op = sbFadeValue[INTERPOLATE16(v1, v2, y & 0xffff)];
		}
#ifdef	MAP_MERGING
	}
#endif

	sbSwap();
}

void
sbApplyMove(void)
{
	unsigned long	*mp;
	unsigned long	*mp0;
	unsigned long	*mp1;
	unsigned char	*op;
	int		i;
	register int	x;
	register int	y;
	int		v1;
	int		tmmf;

	mapsHandleMaps();

	sbCheckAllocation();

	if (mapValid == 0)
	{
		int	size;

		memset(sbData[0], '\0', sbWidth * sbHeight);

		if (WSInit(_width, _height, WS_TITLE_LAYER))
		{
			size = _myHeight / 8;
			WSSetFont(colorMap[255], "verdana", size * 15 / 10, FW_BOLD, WS_TITLE_LAYER);
			WSSupDrawText(_width / 2, _myHeight / 2 - 2 * size, "Smear", WS_TITLE_LAYER, TA_CENTER);
			WSSetFont(colorMap[255], "verdana", size * 9 / 10, FW_BOLD, WS_TITLE_LAYER);
			WSSupDrawText(_width / 2, _myHeight / 2, "mykel :: xplo", WS_TITLE_LAYER, TA_CENTER);

			WSCopyOut(WS_TITLE_LAYER);
		}

		WSCopyHoldToSB(sbData[0], 255, WS_TITLE_LAYER);

		sbSwap();

		return;
	}

#define	INTERPOLATE16(a, b, v)	((a) + ((((b) - (a)) * (v)) >> 16))

#ifdef	MAP_MERGING
	if (mapMerging && mapMergeFactor > 0 && mapMergeFactor < 255)
	{
		mp0 = mapsData[0];
		mp1 = mapsData[1];
		op = sbData[1];

		i = sbWidth * sbHeight;
		tmmf = 256 - mapMergeFactor;

		__asm
		{
loopSpecial:
			mov	edx,	mp1

			mov	eax,	[edx]
			shr	eax,	4
			imul	eax,	tmmf

			mov	ecx,	[edx + 4]
			shr	ecx,	4
			imul	ecx,	tmmf

			mov	edx,	mp0
			mov	ebx,	[edx]
			shr	ebx,	4
			imul	ebx,	mapMergeFactor
			add	eax,	ebx

			mov	ebx,	[edx + 4]
			shr	ebx,	4
			imul	ebx,	mapMergeFactor
			add	ebx,	ecx

			shr	eax,	4
			shr	ebx,	4

			mov	x,	eax
			mov	y,	ebx
			add	mp0,	8
			add	mp1,	8

			mov	ecx,	yCoordinates
			shr	ebx,	14
			and	ebx,	0xfffffffc
			add	ecx,	ebx
			shr	eax,	16
			add	eax,	[ecx]
			mov	edx,	[sbData]
			add	edx,	eax

			xor	eax,	eax
			mov	al,	[edx]
			xor	ebx,	ebx
			mov	bl,	[edx + 1]
			mov	ecx,	x
			and	ecx,	0x0000ffff
			sub	ebx,	eax
			imul	ebx,	ecx
			sar	ebx,	16
			add	eax,	ebx
			mov	v1,	eax

			add	edx,	sbWidth

			xor	eax,	eax
			mov	al,	[edx]
			xor	ebx,	ebx
			mov	bl,	[edx + 1]
			sub	ebx,	eax
			imul	ebx,	ecx
			sar	ebx,	16
			add	eax,	ebx

			mov	ebx,	eax
			mov	eax,	v1
			mov	ecx,	y
			and	ecx,	0x0000ffff
			sub	ebx,	eax
			imul	ebx,	ecx
			sar	ebx,	16
			add	eax,	ebx

			xor	ebx,	ebx
			mov	bl,	[sbFadeValue + eax]
			mov	eax,	op
			mov	[eax],	bl

			inc	op
			dec	i
			jnz	loopSpecial
		}
	}
	else
	{
#endif
		mp = mapsData[0];
		op = sbData[1];

		i = sbWidth * sbHeight;

		__asm
		{
loopNormal:
			mov	ecx,	mp
			mov	eax,	[ecx]
			mov	ebx,	[ecx + 4]
			mov	x,	eax
			mov	y,	ebx
			add	mp,	8

			mov	ecx,	yCoordinates
			shr	ebx,	14
			and	ebx,	0xfffffffc
			add	ecx,	ebx
			shr	eax,	16
			add	eax,	[ecx]
			mov	edx,	[sbData]
			add	edx,	eax

			xor	eax,	eax
			mov	al,	[edx]
			xor	ebx,	ebx
			mov	bl,	[edx + 1]
			mov	ecx,	x
			and	ecx,	0x0000ffff
			sub	ebx,	eax
			imul	ebx,	ecx
			sar	ebx,	16
			add	eax,	ebx
			mov	v1,	eax

			add	edx,	sbWidth

			xor	eax,	eax
			mov	al,	[edx]
			xor	ebx,	ebx
			mov	bl,	[edx + 1]
			sub	ebx,	eax
			imul	ebx,	ecx
			sar	ebx,	16
			add	eax,	ebx

			mov	ebx,	eax
			mov	eax,	v1
			mov	ecx,	y
			and	ecx,	0x0000ffff
			sub	ebx,	eax
			imul	ebx,	ecx
			sar	ebx,	16
			add	eax,	ebx

			xor	ebx,	ebx
			mov	bl,	[sbFadeValue + eax]
			mov	eax,	op
			mov	[eax],	bl

			inc	op
			dec	i
			jnz	loopNormal
		}

#ifdef	MAP_MERGING
	}
#endif

	sbSwap();
}



//////////////////////////////////////////////////////////////////////////////////////
//
// cmap (Colour Map)
//
// Functions for dealing with the colour map
//
//////////////////////////////////////////////////////////////////////////////////////


void
cmapMapSBToVideo(void)
{
	unsigned long	*op;
	unsigned long	*lop;
	unsigned char	*ip;
	int				x;
	int				y;

	op = _Video;
	ip = sbData[0];

	for (y = 0; y < _Ymin; y++, op += _pitch)
		memset(op, 0, sizeof(unsigned long) * _width);

	for (y = 0; y < _myHeight; y++, op += _pitch)
		for (lop = op, x = 0; x < _width; x++, lop++, ip++)
			*lop = colorMap[*ip];

	for (y = _Ymax; y < _height; y++, op += _pitch)
		memset(op, 0, sizeof(unsigned long) * _width);
}



//////////////////////////////////////////////////////////////////////////////////////
//
// actually do something here.
//
//////////////////////////////////////////////////////////////////////////////////////
void
effectsDrawLine(int x1, int y1, int x2, int y2, int color, int blend)
{
	int	i;
	int	x, y;
	int	stx;
	int	sty;
	int	steps;

	if ((stx = x2 - x1) < 0)
		stx = - stx;

	if ((sty = y2 - y1) < 0)
		sty = - sty;

	if (stx > sty)
		steps = stx;
	else
		steps = sty;

	if (steps == 0)
		return;

	stx = ((x2 - x1) << 16) / steps;
	sty = ((y2 - y1) << 16) / steps;

	x1 <<= 16;
	y1 <<= 16;

	for (i = 0; i < steps; i++)
	{
		x = x1 >> 16;
		y = y1 >> 16;
		if (x >= 1 && x < _width - 1 && y >= 1 && y < _myHeight - 1)
			switch (blend)
			{
			case 1:
				if ((short)(pxlXYPixel(sbData[0], x, y)) + color > 255)
					pxlXYPixel(sbData[0], x, y) = 255;
				else
					pxlXYPixel(sbData[0], x, y) += color;
				break;

			case 2:
				if ((short)(pxlXYPixel(sbData[0], x, y)) < color)
					pxlXYPixel(sbData[0], x, y) = color;

				break;
			default:
				pxlXYPixel(sbData[0], x, y) = color;
			}

		x1 += stx;
		y1 += sty;
	}
}

typedef	struct	_xycl
{
	short	x[512];
	short	y[512];
	short	color[512];
	short	last;
}	XYCL;

XYCL	xycls[2][MAX_WAVEFORM_PATHS];

void
effectsInitBuffer(void)
{
	memset(xycls, 0, sizeof(xycls));
}

void
effectsBufferLine(int x, int y, int color, int whichDraw, int path, int pos)
{
	XYCL	*xyclp = &(xycls[whichDraw - 1][path]);
	int	i;

	xyclp->x[pos] = x;
	xyclp->y[pos] = y;
	xyclp->color[pos] = color;

	if (pos == 0)
	{
		xyclp->last = pos;
		return;
	}

	for (i = xyclp->last + 1; i < pos; i++)
	{
		// Interpolate X/Y
		xyclp->x[i] = xyclp->x[xyclp->last] +
			((xyclp->x[pos] - xyclp->x[xyclp->last]) *
			 (i - xyclp->last) / (pos - xyclp->last));
		xyclp->y[i] = xyclp->y[xyclp->last] +
			((xyclp->y[pos] - xyclp->y[xyclp->last]) *
			 (i - xyclp->last) / (pos - xyclp->last));
		xyclp->color[i] = xyclp->color[xyclp->last] +
			((xyclp->color[pos] - xyclp->color[xyclp->last]) *
			 (i - xyclp->last) / (pos - xyclp->last));
	}

	xyclp->last = pos;
}


effectsDrawMerge(int n0, int n1, int v0, int v1, int up)
{
	XYCL	*l0 = &(xycls[0][n0]);
	XYCL	*l1 = &(xycls[1][n1]);
	int	i;
	int	m = l0->last;

	if (l1->last < m)
		m = l1->last;

#define	VBLEND(a, b) ((((a) * v0) + ((b) * v1)) >> 8)
	if (up)
		for (i = 0; i < m; i++)
			effectsDrawPoint(VBLEND(l0->x[i]    , l1->x[i]    ),
					 VBLEND(l0->y[i]    , l1->y[i]    ),
					 VBLEND(l0->color[i], l1->color[i]), 0);
	else
		for (i = 1; i < m; i++)
			effectsDrawLine(VBLEND(l0->x[i - 1], l1->x[i - 1]),
					VBLEND(l0->y[i - 1], l1->y[i - 1]),
					VBLEND(l0->x[i]    , l1->x[i]    ),
					VBLEND(l0->y[i]    , l1->y[i]    ),
					VBLEND(l0->color[i], l1->color[i]), 0);
#undef	VBLEND
}



effectsDrawBuffer(int np0, int np1, int *up0, int *up1, int v0, int v1)
{
	int	count = np0;
	int	i;

	if (count < np1)
		count = np1;

	for (i = 0; i < count; i++)
		effectsDrawMerge(i % np0, i % np1, v0, v1, (up0[i % np0] || up1[i % np1]));
}



void
effectsDrawPoint(int x, int y, int color, int blend)
{
	if (x >= 1 && x < _width - 1 && y >= 1 && y < _myHeight - 1)
		switch (blend)
		{
		case 1:
			if ((short)(pxlXYPixel(sbData[0], x, y)) + color > 255)
				pxlXYPixel(sbData[0], x, y) = 255;
			else
				pxlXYPixel(sbData[0], x, y) += color;
			break;

		case 2:
			if ((short)(pxlXYPixel(sbData[0], x, y)) < color)
				pxlXYPixel(sbData[0], x, y) = color;

			break;
		default:
			pxlXYPixel(sbData[0], x, y) = color;
		}
}

void
effectsDrawDirectPoint(int x, int y, unsigned long color)
{
	if (x >= 0 && x < _width  && y >= 0 && y < _myHeight)
	{
		y += (_height - _myHeight) >> 1;
		if (_Video)
		{
//#define _s_ARGB(a, r, g, b) (RGB(r,g,b) | ((BYTE)a<<24))
			//_Video[(y*_pitch)+x] = _s_ARGB(255, b, g, r);
			_Video[(y*_pitch)+x] = (unsigned long)color;
		}
//#undef _s_ARGB
	}

}

void
effectsDrawParticle(int x, int y, int xe, int ye, int h, int w, int color, int style, int blend)
{
	int	dif;
	int	i, j, x0, y0;

	if (style == 2)
	{
		effectsDrawLine(x, y, xe, ye, color, blend);
		return;
	}

	if (h < 16 || w < 16)
		return;

	if (h < 256)
	{
		color = (color * h) >> 8;
		h = 256;
	}

	if (w < 256)
	{
		color = (color * w) >> 8;
		w = 256;
	}

	h >>= 8;
	w >>= 8;

	dif = (h + w) / 2;

	for (y0 = y - h / 2, i = 0; i <= h; i++, y0++)
		for (x0 = x - w / 2, j = 0; j <= w; j++, x0++)
			if ((x0 - x) * (x0 - x) + (y0 - y) * (y0 - y) < dif)
				effectsDrawPoint(x0, y0, color, blend);

}


static char		*clickedWaveformName;

void
waveformHandle(void)
{
static	int		waveformChanging = 0;
static	int		waveformStartTime = 0;
static	int		waveformEndTime = 0;
static	Environment	*waveformEnvironments[2] = { NULL, NULL };

	int	now;

	now = GetTickCount();

	if (ssmCheckNextWaveform())
		waveformEndTime = now;

	if (waveformEnvironments[0] == (Environment *)0)
	{

		waveformEnvironments[0] = ChooseWaveFormScript();
		SetupForWaveForm(waveformEnvironments[0]);
		clickedWaveformName = waveformEnvironments[0]->eFile;

		waveformChanging = 0;
		waveformStartTime = now;
		if (ARG_TotalMayhem)
			waveformEndTime = waveformStartTime + 250;
		else
			waveformEndTime = waveformStartTime + 100 * ARG_WaveformHoldTime * (SIMPLE_RAND(100) + 50) / 100;
	}

	if (now >= waveformEndTime)
	{
		waveformStartTime = now;
		waveformChanging = !waveformChanging;

		if (waveformChanging)
		{
			waveformEnvironments[1] = ChooseWaveFormScript();
			SetupForWaveForm(waveformEnvironments[1]);

			waveformEndTime = waveformStartTime + 100 * ARG_WaveformFadeTime * (SIMPLE_RAND(100) + 50) / 100;
		}
		else
		{
			waveformEnvironments[0] = waveformEnvironments[1];
			clickedWaveformName = waveformEnvironments[0]->eFile;
			if (ARG_TotalMayhem)
				waveformEndTime = waveformStartTime + 250;
			else
				waveformEndTime = waveformStartTime + 100 * ARG_WaveformHoldTime * (SIMPLE_RAND(100) + 50) / 100;
		}
	}

	if (waveformChanging)
	{
		int v1 = (now - waveformStartTime) * 256 / (waveformEndTime - waveformStartTime);
		int v0 = 256 - v1;

		effectsInitBuffer();

		RunForWaveForm(waveformEnvironments[0], _width, _myHeight, 1);
		RunForWaveForm(waveformEnvironments[1], _width, _myHeight, 2);

		effectsDrawBuffer(waveformEnvironments[0]->eRuntime.reWaveform.numPaths,
				  waveformEnvironments[1]->eRuntime.reWaveform.numPaths,
				  waveformEnvironments[0]->eRuntime.reWaveform.usePoints,
				  waveformEnvironments[1]->eRuntime.reWaveform.usePoints,
				  v0, v1);
	}
	else
		RunForWaveForm(waveformEnvironments[0], _width, _myHeight, 0);
}


void
particleHandle(void)
{
static	int		numRunning = 0;
static	Environment	*particleEnvironments[MAX_PARTICLES_RUNNING];
static	int		particleState[MAX_PARTICLES_RUNNING];
static	int		particleStartTime[MAX_PARTICLES_RUNNING];
static	int		particleEndTime[MAX_PARTICLES_RUNNING];
static	int		__inited = 0;
	double		fade = 0.0;
	int		i;
static	int		lastTry;

	int		now;
	int		forceNew = 0;

	if (ssmCheckNextParticle())
	{
		forceNew = 1;
		numRunning = 0;
		for (i = 0; i < MAX_PARTICLES_RUNNING; i++)
			particleState[i] = 0;
	}

	now = GetTickCount();

	if (__inited)
	{
		__inited = 1;
		memset(particleEnvironments, 0, sizeof(particleEnvironments));
		memset(particleState, 0, sizeof(particleState));
		memset(particleStartTime, 0, sizeof(particleStartTime));
		memset(particleEndTime, 0, sizeof(particleEndTime));
	}

	if (ARG_TotalMayhem && numRunning == 0)
		forceNew = 1;

	// Start new ones
	if (forceNew || (numRunning < MAX_PARTICLES_RUNNING &&
	    lastTry + PARTICLE_TRY_TIMEOUT < now))
	{
		double	d = (double)SIMPLE_RAND(32767) / 32767.0;
		int	yes = 0;
		double	c = 1.0;

		lastTry = now;

		for (i = 0; !yes && i < MAX_PARTICLES_RUNNING; i++)
		{
			if (numRunning > i)
				continue;

			c *= VALUE_ParticleChance;

			if (d >= c)
				continue;

			yes = 1;
		}

		if (forceNew || yes)
		{
			for (i = 0; i < MAX_PARTICLES_RUNNING; i++)
				if (particleState[i] == 0)
					break;

			if (i < MAX_PARTICLES_RUNNING &&
			    (particleEnvironments[i] = ChooseParticleScript()) != (Environment *)0)
			{
				int	j;

				for (j = 0; j < MAX_PARTICLES_RUNNING; j++)
					if (particleState[j] > 0 &&
					    particleEnvironments[j] ==
					    particleEnvironments[i])
						j = 9999;

				if (j < 9999)
				{
					SetupForParticle(particleEnvironments[i]);
					particleState[i] = 1;
					particleStartTime[i] = now;
					particleEndTime[i] = now + 100 * ARG_ParticleFadeTime * (SIMPLE_RAND(100) + 50) / 100;

					numRunning++;
				}
			}

		}
	}

	for (i = 0; i < MAX_PARTICLES_RUNNING; i++)
	{
		if (particleState[i] == 0)
			continue;

		if (now > particleEndTime[i])
		{
			particleState[i]++;

			if (particleState[i] > 3)
			{
				particleState[i] = 0;
				numRunning--;
				continue;
			}

			particleStartTime[i] = now;

			if (particleState[i] == 2)
			{
				if (ARG_TotalMayhem)
					particleEndTime[i] = now + 250;
				else
					particleEndTime[i] = now + 100 * ARG_ParticleHoldTime * (SIMPLE_RAND(100) + 50) / 100;
			}
			else
				particleEndTime[i] = now + 100 * ARG_ParticleFadeTime * (SIMPLE_RAND(100) + 50) / 100;
		}

		switch (particleState[i])
		{
		case 1:
			fade = (double)(particleEndTime[i] - now) / (double)(particleEndTime[i] - particleStartTime[i]);
			break;

		case 2:
			fade = 0.0;
			break;

		case 3:
			fade = (double)(now - particleStartTime[i]) / (double)(particleEndTime[i] - particleStartTime[i]);
			break;

		default:
			fade = 1.0;
			break;
		}

		if (fade < 0.0)
			fade = 0.0;

		if (fade > 1.0)
			fade = 1.0;


		RunForParticle(particleEnvironments[i], _width, _myHeight, fade);
	}
}

static	int	songNameChanged = SONG_TOTAL_TIME + 1;
static	char	songName[256] = "";
static	char	songArtist[256] = "";
static	int		effectMade	= 1;
static	int		initeddev	= 0;
static	unsigned long *	TextPreDraw = 0;

void
handleSongName(void)
{
	int	size = _myHeight / 6;
	int	color;
static	int	docheck = 0;
extern	void	C_CheckSongName(char *, char *, int *);

	if (!ARG_ShowTitles)
		return;

	if (docheck++ > 50)
	{
		docheck = 0;
		C_CheckSongName(songName, songArtist, &songNameChanged);
		if (songNameChanged == 0)
			WSDeInit(WS_SONGNAME_LAYER);
	}

	if (songNameChanged > SONG_TOTAL_TIME && effectMade)
		return;

	color = 255;

	if (WSInit(_width, _height, WS_SONGNAME_LAYER))
	{
		size = _myHeight / 12;

		WSSetFont(colorMap[color], "verdana", size * 14 / 10, FW_BOLD, WS_SONGNAME_LAYER);
		WSSupDrawText(5,5,songArtist, WS_SONGNAME_LAYER, TA_LEFT);
		WSSetFont(colorMap[color], "verdana", size * 11 / 10, FW_BOLD, WS_SONGNAME_LAYER);
		WSSupDrawText(5,5 + size * 2,songName, WS_SONGNAME_LAYER, TA_LEFT);

		WSCopyOut(WS_SONGNAME_LAYER);
	}

	
	if (songNameChanged > SONG_TOTAL_TIME)
	{
		WSCopyHoldToSB(sbData[0], color, WS_SONGNAME_LAYER);

		WSDeInit(WS_SONGNAME_LAYER);

		effectMade = 1;

		return;
	}


	WSCopyHoldToVid(_Video, _pitch, (_height - _myHeight) / 2, colorMap[color], WS_SONGNAME_LAYER);
	// WSCopyHoldToSB(sbData[0], 127, WS_SONGNAME_LAYER);
	
	effectMade = 0;
	songNameChanged++;
}


void
effectsHandle(void)
{
	waveformHandle();
	particleHandle();
	// handleSongName();
}


//////////////////////////////////////////////////////////////////////////////////////
//
// visInit(void)
//
// Called before plugin starts.  Use it to init things.  Also called after a DeInit,
// when the plugin needs to restart, if the DeInit returned TRUE
//
//////////////////////////////////////////////////////////////////////////////////////
void
visInit(void)
{
	static	int	_done_init = 0;
	
#ifdef	USE_MAP_THREADS
	if (ARG_ThreadMaps)
		threadSetupSubThread();
#endif

	if (_done_init)
	{
#ifdef	USE_MAP_THREADS
		if (ARG_ThreadMaps)
			mapRestart = 1;
#endif
		return;
	}
	
	_done_init = 1;
	
	srand(GetTickCount());

	LoadEverything();
	ssmSetupMemory();

	return;
}

static char *clickedCMAPName;

void
colormapHandle(void)
{
static	int		colormapChanging = 0;
static	int		colormapStartTime = 0;
static	int		colormapEndTime = 0;
static	Environment	*colormapEnvironments[2] = { NULL, NULL };

	int	now;
	int	forceRun = 0;

	now = GetTickCount();

	if (ssmCheckNextColor())
		colormapEndTime = now;


	if (colormapEnvironments[0] == (Environment *)0)
	{
		colormapEnvironments[0] = ChooseColorScript();
		SetupForColorRun(colormapEnvironments[0]);
		clickedCMAPName = colormapEnvironments[0]->eFile;

		colormapChanging = 0;
		colormapStartTime = now;
		if (ARG_TotalMayhem)
			colormapEndTime = colormapStartTime + 250;
		else
			colormapEndTime = colormapStartTime + 100 * ARG_ColormapHoldTime * (SIMPLE_RAND(100) + 50) / 100;

		forceRun = 1;
	}

	if (now >= colormapEndTime)
	{
		colormapStartTime = now;
		colormapChanging = !colormapChanging;
		forceRun = 1;

		if (colormapChanging)
		{
			colormapEnvironments[1] = ChooseColorScript();
			SetupForColorRun(colormapEnvironments[1]);

			colormapEndTime = colormapStartTime + 100 * ARG_ColormapFadeTime * (SIMPLE_RAND(100) + 50) / 100;
		}
		else
		{
			colormapEnvironments[0] = colormapEnvironments[1];
			clickedCMAPName = colormapEnvironments[0]->eFile;
			if (ARG_TotalMayhem)
				colormapEndTime = colormapStartTime + 250;
			else
				colormapEndTime = colormapStartTime + 100 * ARG_ColormapHoldTime * (SIMPLE_RAND(100) + 50) / 100;
		}
	}

	if (colormapChanging)
	{
		unsigned char	*i0p, *i1p, *op;
		int		v0, v1;
		int		i;

		RunForColor(colormapEnvironments[0], fadeMaps[0], forceRun);
		RunForColor(colormapEnvironments[1], fadeMaps[1], forceRun);

		op = (char *)colorMap;
		i0p = (char *)fadeMaps[0];
		i1p = (char *)fadeMaps[1];
		v1 = (now - colormapStartTime) * 256 / (colormapEndTime - colormapStartTime);
		if (v1 < 0)
			v1 = 0;
		if (v1 > 256)
			v1 = 256;
		v0 = 256 - v1;
		for (i = 0; i < 256; i++)
		{
			*op = (unsigned char)(((int)(*i0p) * v0 + (int)(*i1p) * v1) >> 8);
			op++; i0p++; i1p++;
			*op = (unsigned char)(((int)(*i0p) * v0 + (int)(*i1p) * v1) >> 8);
			op++; i0p++; i1p++;
			*op = (unsigned char)(((int)(*i0p) * v0 + (int)(*i1p) * v1) >> 8);
			op++; i0p++; i1p++;
			op++; i0p++; i1p++;
		}
	}
	else
		RunForColor(colormapEnvironments[0], colorMap, forceRun);
}


static int	clickedRunner = 0;

void
handleClicked(void)
{
	if (clickedRunner == 0)
		return;

	if (WSInit(_width, _height, WS_CLICKED_LAYER))
	{
		int	size = _myHeight / 12;
		char	_tmp[256];

		size = _myHeight / 8;
		WSSetFont(colorMap[255], "verdana", size * 9 / 10, FW_BOLD, WS_CLICKED_LAYER);
		WSSupDrawText(5,
			      _myHeight - (4 * (size + 2)) - 2,
			      "Current Scripts:", 
			      WS_CLICKED_LAYER, 
			      TA_LEFT);
		sprintf(_tmp, "Color: %.*s", strlen(clickedCMAPName) - 4, clickedCMAPName);
		WSSupDrawText(5,
			      _myHeight - (3 * (size + 2)) - 2,
			      _tmp, 
			      WS_CLICKED_LAYER, 
			      TA_LEFT);
		//WSSetFont(colorMap[255], "verdana", size * 9 / 10, FW_BOLD, WS_CLICKED_LAYER);
		sprintf(_tmp, "Wave: %.*s", strlen(clickedWaveformName) - 4, clickedWaveformName);
		WSSupDrawText(5,
			      _myHeight - (2 * (size + 2)) - 2,
			      _tmp, 
			      WS_CLICKED_LAYER, 
			      TA_LEFT);
		//WSSetFont(colorMap[255], "verdana", size * 9 / 10, FW_BOLD, WS_CLICKED_LAYER);
		sprintf(_tmp, "Move: %.*s", strlen(clickedMoveName) - 4, clickedMoveName);
		WSSupDrawText(5,
			      _myHeight - (1 * (size + 2)) - 2, 
			      _tmp, 
			      WS_CLICKED_LAYER,
			      TA_LEFT);

		WSCopyOut(WS_CLICKED_LAYER);
	}

	WSCopyHoldToVid(_Video, _pitch, (_height - _myHeight) / 2, colorMap[255], WS_CLICKED_LAYER);
	
	clickedRunner--;
	
	if (clickedRunner == 0)
	{
		WSCopyHoldToSB(sbData[0], 255, WS_CLICKED_LAYER);
		WSDeInit(WS_CLICKED_LAYER);
	}

}



//////////////////////////////////////////////////////////////////////////////////////
//
// visRender(unsigned long *Video, int width, int height, int pitch, VisData *pVD)
//
// This is the main Render function for your vis.  It is called whenever sonique wants
// a new frame.
//
// Draw your vis information.  The current Video data is the blur/smoke/zoom effected
// frame that will be output after you modify it.  You should only clear the whole
// frame if you do not want blur/smoke/zoom to work.
//
// It is important to stay within the memory use defined by width/height/pitch
//
// The point (x, y) can be referenced as follows
//
//    ASSET(x >= 0 && x < width && y >= 0 && y < height);
//    Video[y * pitch + x] = VALUE
//
// Note: pitch most probably is not equal to width, and is often negative.
//
// Return TRUE when completed successfully.
//
//////////////////////////////////////////////////////////////////////////////////////
BOOL
visRender(unsigned long* Video, int width, int height, int pitch, VisData* pVD)
{
static	int	which = 0;
//static	int	timing = 0;

	_Video	= Video;
	_width	= width;
	_height	= height;
	_pitch	= pitch;
	_pVD	= pVD;
	_div	= (width + height) >> 1;


	_myHeight = (width * ARG_LetterBoxPercent) / 100;
	if (_myHeight > height)
		_myHeight = height;
	_Ymin = (height - _myHeight) / 2;
	_Ymax = _Ymin + _myHeight;

	CalculateAudioData(_pVD->Waveform[0], _pVD->Waveform[1],
			   _pVD->Spectrum[0], _pVD->Spectrum[1]);

	pxlSetupYCoordinates(height, width);

	colormapHandle();

#if	0
	if (timing == 0)
	{
		long	l, t, t2;
		char	_tmp[512];

		timing = 1;

		t = GetTickCount();

		for (l = 0; l < 250; l++)
			sbApplyMove();

		t = GetTickCount() - t;

		t2 = GetTickCount();

		for (l = 0; l < 250; l++)
			OLDsbApplyMove();

		t2 = GetTickCount() - t2;

		sprintf(_tmp, "Standard\nper iteration\nNEW: %.2lf ms\nOLD: %.2lf ms", 
			(double)t / 250.0,
			(double)t2 / 250.0);
		MessageBox(NULL, _tmp, "Timing",
		   MB_ICONWARNING | MB_APPLMODAL | MB_OK | MB_SETFOREGROUND);
	}

	if (timing == 1 && mapMerging && mapMergeFactor > 0 && mapMergeFactor < 255)
	{
		long	l, t, t2;
		char	_tmp[512];

		timing = 2;

		t = GetTickCount();

		for (l = 0; l < 250; l++)
			sbApplyMove();

		t = GetTickCount() - t;

		t2 = GetTickCount();

		for (l = 0; l < 250; l++)
			OLDsbApplyMove();

		t2 = GetTickCount() - t2;

		sprintf(_tmp, "Merging\nper iteration\nNEW: %.2lf ms\nOLD: %.2lf ms", 
			(double)t / 250.0,
			(double)t2 / 250.0);
		MessageBox(NULL, _tmp, "Timing",
		   MB_ICONWARNING | MB_APPLMODAL | MB_OK | MB_SETFOREGROUND);
	}
#endif

	sbApplyMove();
	effectsHandle();

	cmapMapSBToVideo();
	handleSongName();
	handleClicked();

	return TRUE;
}



//////////////////////////////////////////////////////////////////////////////////////
//
// visSaveSettings(char *FileName)
//
// Called just before the vis is closed (probably when sonique is closed).  Use it to
// save configuration options to the VIS.INI file.  There are some inline functions to
// help with saving integer and boolean settings.
//
//////////////////////////////////////////////////////////////////////////////////////
__inline	void
__SaveIntSetting(char *FileName, char *name, int setting, int newv)
{
	char	Temp[64];

	if (!newv)
		return;

	sprintf(Temp, "%d", setting);
	WritePrivateProfileString(VIS_SHORT_NAME, name, Temp, FileName);
}


__inline	void
__SaveTFSetting(char *FileName, char *name, int setting, int newv)
{
	char	Temp[64];

	if (!newv)
		return;

	sprintf(Temp, "%s", setting ? "True" : "False");
	WritePrivateProfileString(VIS_SHORT_NAME, name, Temp, FileName);
}


BOOL
visSaveSettings(char* FileName)
{
//	__SaveTFSetting(FileName, "Sample Boolean (True-False)", ARG_SampleTF, WRITE_SampleTF);
//	__SaveIntSetting(FileName, "Sample Integer (1-100)", ARG_SampleInt, WRITE_SampleInt);
	__SaveIntSetting(FileName, "Movemap Hold Time (1/10ths of seconds)", ARG_MovemapHoldTime, WRITE_MovemapHoldTime);
#ifdef	MAP_MERGING
	__SaveIntSetting(FileName, "Movemap Fade Time (1/10ths of seconds)", ARG_MovemapFadeTime, WRITE_MovemapFadeTime);
#endif
	__SaveIntSetting(FileName, "Colormap Hold Time (1/10ths of seconds)", ARG_ColormapHoldTime, WRITE_ColormapHoldTime);
	__SaveIntSetting(FileName, "Colormap Fade Time (1/10ths of seconds)", ARG_ColormapFadeTime, WRITE_ColormapFadeTime);
	__SaveIntSetting(FileName, "Waveform Hold Time (1/10ths of seconds)", ARG_WaveformHoldTime, WRITE_WaveformHoldTime);
	__SaveIntSetting(FileName, "Waveform Fade Time (1/10ths of seconds)", ARG_WaveformFadeTime, WRITE_WaveformFadeTime);
	__SaveIntSetting(FileName, "Particle Hold Time (1/10ths of seconds)", ARG_ParticleHoldTime, WRITE_ParticleHoldTime);
	__SaveIntSetting(FileName, "Particle Fade Time (1/10ths of seconds)", ARG_ParticleFadeTime, WRITE_ParticleFadeTime);
	__SaveIntSetting(FileName, "Particle Chance per Frame (0-100)", ARG_ParticleChance, WRITE_ParticleChance);
	__SaveIntSetting(FileName, "Letter Box Percentage (10-100)", ARG_LetterBoxPercent, WRITE_LetterBoxPercent);
	__SaveTFSetting(FileName, "Show Titles (True-False)", ARG_ShowTitles, WRITE_ShowTitles);
	__SaveTFSetting(FileName, "Debug Scripts (True-False)", ARG_DebugScripts, WRITE_DebugScripts);
	__SaveTFSetting(FileName, "Disable Smear Title Screen (True-False)", ARG_NoVisName, WRITE_NoVisName);
#ifdef	USE_MAP_THREADS
	__SaveTFSetting(FileName, "Thread Move Maps (True-False)", ARG_ThreadMaps, WRITE_ThreadMaps);
#endif
#ifdef	MAP_SUPERMERGE
	__SaveTFSetting(FileName, "Map Constant Merge (True-False)", ARG_MovemapConstantFade, WRITE_MovemapConstantFade);
	__SaveTFSetting(FileName, "Total Mayhem (True-False)", ARG_TotalMayhem, WRITE_TotalMayhem);
#endif

	WritePrivateProfileString(VIS_SHORT_NAME, "Version", VIS_VERSION, FileName);
	WritePrivateProfileString(VIS_SHORT_NAME, "Author", VIS_AUTHOR, FileName);
	
	return TRUE;
}



//////////////////////////////////////////////////////////////////////////////////////
//
// visOpenSettings(char *FileName)
//
// Called when the vis is first loaded, before the visInit() function.  Use it to load
// configuration options from the VIS.INI file.  There are sone inline functions to
// help with loading integer and boolean settings.
//
//////////////////////////////////////////////////////////////////////////////////////
__inline	void
__LoadIntSetting(char *FileName, char *name, int *settingp, int *newp, int mini, int maxi, int defa)
{
	char	TestBuffer[64];

	GetPrivateProfileString(VIS_SHORT_NAME, name, "9999", TestBuffer, sizeof(TestBuffer), FileName);
	*newp = strcmp(TestBuffer, "9999") == 0;
	if (*newp)
		*settingp = defa;
	else
		*settingp = atoi(TestBuffer);

	if (*settingp < mini)
		*settingp = mini;

	if (*settingp > maxi)
		*settingp = maxi;
}

__inline	void
__LoadTFSetting(char *FileName, char *name, int *settingp, int *newp, int defa)
{
	char	TestBuffer[64];

	GetPrivateProfileString(VIS_SHORT_NAME, name, "9999", TestBuffer, sizeof(TestBuffer), FileName);
	*newp = strcmp(TestBuffer, "9999") == 0;
	if (*newp)
		*settingp = defa;
	else
	{
		if (TestBuffer[0] == 'f' || TestBuffer[0] == 'F' ||
			((TestBuffer[0] == 'O' || TestBuffer[0] == 'o') && (TestBuffer[1] == 'f' || TestBuffer[1] == 'F')) ||
			(TestBuffer[0] == '0' && atoi(TestBuffer) == 0))
				*settingp = 0;
		else
			*settingp = 1;
	}
}

BOOL
visOpenSettings(char* FileName)
{
//	__LoadTFSetting(FileName, "Sample Boolean (True-False)", &ARG_SampleTF, &WRITE_SampleTF, 1);
//	__LoadIntSetting(FileName, "Sample Integer (1-100)", &ARG_SampleInt, &WRITE_SampleInt, 1, 100, 50);
	__LoadIntSetting(FileName, "Movemap Hold Time (1/10ths of seconds)", &ARG_MovemapHoldTime, &WRITE_MovemapHoldTime, 10, 36000, 150);
#ifdef	MAP_MERGING
	__LoadIntSetting(FileName, "Movemap Fade Time (1/10ths of seconds)", &ARG_MovemapFadeTime, &WRITE_MovemapFadeTime, 10, 36000, 30);
#endif
	__LoadIntSetting(FileName, "Colormap Hold Time (1/10ths of seconds)", &ARG_ColormapHoldTime, &WRITE_ColormapHoldTime, 10, 36000, 130);
	__LoadIntSetting(FileName, "Colormap Fade Time (1/10ths of seconds)", &ARG_ColormapFadeTime, &WRITE_ColormapFadeTime, 10, 6000, 30);
	__LoadIntSetting(FileName, "Waveform Hold Time (1/10ths of seconds)", &ARG_WaveformHoldTime, &WRITE_WaveformHoldTime, 10, 36000, 150);
	__LoadIntSetting(FileName, "Waveform Fade Time (1/10ths of seconds)", &ARG_WaveformFadeTime, &WRITE_WaveformFadeTime, 10, 6000, 15);
	__LoadIntSetting(FileName, "Particle Hold Time (1/10ths of seconds)", &ARG_ParticleHoldTime, &WRITE_ParticleHoldTime, 10, 36000, 100);
	__LoadIntSetting(FileName, "Particle Fade Time (1/10ths of seconds)", &ARG_ParticleFadeTime, &WRITE_ParticleFadeTime, 10, 36000, 30);
	__LoadIntSetting(FileName, "Particle Chance per Frame (0-100)", &ARG_ParticleChance, &WRITE_ParticleChance, 0, 100, 10);
	__LoadIntSetting(FileName, "Letter Box Percentage (10-100)", &ARG_LetterBoxPercent, &WRITE_LetterBoxPercent, 10, 100, 100);
	__LoadTFSetting(FileName, "Show Titles (True-False)", &ARG_ShowTitles, &WRITE_ShowTitles, 1);
	__LoadTFSetting(FileName, "Debug Scripts (True-False)", &ARG_DebugScripts, &WRITE_DebugScripts, 0);
	__LoadTFSetting(FileName, "Disable Smear Title Screen (True-False)", &ARG_NoVisName, &WRITE_NoVisName, 0);
#ifdef	USE_MAP_THREADS
	__LoadTFSetting(FileName, "Thread Move Maps (True-False)", &ARG_ThreadMaps, &WRITE_ThreadMaps, 1);
#endif
#ifdef	MAP_SUPERMERGE
	__LoadTFSetting(FileName, "Map Constant Merge (True-False)", &ARG_MovemapConstantFade, &WRITE_MovemapConstantFade, 0);
	__LoadTFSetting(FileName, "Total Mayhem (True-False)", &ARG_TotalMayhem, &WRITE_TotalMayhem, 0);
	if (ARG_TotalMayhem)
	{
		WRITE_MovemapConstantFade = (ARG_MovemapConstantFade != 1);
		ARG_MovemapConstantFade = 1;
	}

	if (ARG_MovemapConstantFade)
	{
		WRITE_NoVisName = (ARG_NoVisName != 1);
		ARG_NoVisName = 1;
	}
#endif

	VALUE_ParticleChance = (double)ARG_ParticleChance / 100.0;
 
	return TRUE;
}



//////////////////////////////////////////////////////////////////////////////////////
//
// visDeInit(void)
//
// Called when the vis is about to be deselected.  This is used to free up resources
// when the vis is not going to be used for a while.  If the function returns TRUE
// then the next time the vis is required, the visInit() function will be called again.
// 
// Only available in API Versions >= 1
//
//////////////////////////////////////////////////////////////////////////////////////
BOOL
visDeInit(void)
{
#ifdef	USE_MAP_THREADS
	if (ARG_ThreadMaps)
		parentShutdownThread();
#endif

	return TRUE;
}



//////////////////////////////////////////////////////////////////////////////////////
//
// visDummyReceiver(void *p)
//
// This is a dummy function that would normally be a QueryReceiver function in a C++
// file.  As the argument passed to this function is a C++ class, it cannot be used
// within the scope of a C program.  If you wish to use the QueryClass functions, you
// will need to use a C++ version of your vis.
//
// Only available in API Versions >= 2
//
//////////////////////////////////////////////////////////////////////////////////////
BOOL
visDummyReceiver(void *p)
{
extern	void	C_SetupQuery(void *);

	C_SetupQuery(p);

	return FALSE;
}



//////////////////////////////////////////////////////////////////////////////////////
//
// visOpenSettings(char *FileName)
//
// Called when the mouse is clicked on the vis window in 'vis mode'.  Currently this
// only happens if the vis is in the normal (small) size on this mode.  Fullwindow
// and midmode are currently not supported.  Also, the buttons value is meaningless
// as it is only triggered by a left mouse click.
//
// Only available in API Versions >= 2
//
//////////////////////////////////////////////////////////////////////////////////////
BOOL
visClicked(int x, int y, int buttons)
{
	if (clickedRunner > 1)
		clickedRunner = 1;
	else
		clickedRunner = CLICKED_LIFETIME;
	
	return TRUE;
}



//////////////////////////////////////////////////////////////////////////////////////
//
// VisInfo __vis_information
//
// This defines the vis properties.  The values are
//
//	 1. Version - the vis api version number (currently 2)
//	 2. Plug-In Name (char*, the name you want displayed when your plug-in is playing)
//	 3. Options (VI_WAVEFORM - if you need the waveform PCM data
//	             VI_SPECTRUM - if you need the spectrum FFT data
//               SONIQUEVISPROC - if you want user defined blurs to effect your plugin)
//	 4. Name of the Initialize function (leave as is)
//	 5. Name of the Render function (leave as is)
//	 6. Name of the SaveSettings function (leave as is)
//	 7. Name of the OpenSettings function (leave as is)
//	 8. Name of the DeInit function (leave as is) (only for version >= 1)
//	 9. Name of the Clicked function (leave as is) (only for version >= 2)
//	10. Name of the QueryReceiver function (leave as is) (only for version >= 2)
//
// You should not need to change anything other than the Options portion of this
// structure.
//
//////////////////////////////////////////////////////////////////////////////////////
VisInfo __vis_information = 
{
	2,
	VIS_LONG_NAME,
	VI_SPECTRUM|VI_WAVEFORM,
	visInit,
	visRender,
	visSaveSettings,
	visOpenSettings,
	visDeInit,
	visClicked,
	visDummyReceiver
};



//////////////////////////////////////////////////////////////////////////////////////
//
// QueryModule(void)
//
// This is the main access function for loading the vis.  It is the first function
// called in the DLL and should return a pointer to a VisInfo structure defining the
// properties of the vis (see above).  You should not need to change this function.
//
//////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT VisInfo	*
QueryModule(void)
{
	return &__vis_information;
}
