#include "smear.h"

#ifdef	USE_MAP_THREADS
//////////////////////////////////////////////////////////////////////
//
// This is the code for handling the sub thread
//
//////////////////////////////////////////////////////////////////////

static	HANDLE	subThread;
static	HANDLE	subMutex;
static	HANDLE	subEventChildToParent;
static	HANDLE	subEventACKChildToParent;
static	HANDLE	subEventParentToChild;
static	HANDLE	subEventACKParentToChild;

#define	P2C_CLEAR				0x00
#define	P2C_START_BUILD			0x01
#define	P2C_START_FAST_BUILD	0x02
#define	P2C_STOP_BUILD			0x03
#define	P2C_EXIT_THREAD			0x04

#define	C2P_CLEAR				0x00
#define	C2P_FINISHED			0x01
#define	C2P_INCOMPLETE			0x02

static	int		msgParentToChild = P2C_CLEAR;
static	int		msgChildToParent = C2P_CLEAR;

static	int				childExiting = 0;
static	int				childBuilding = 0;
static	int				parentBuilding = 0;
static	DWORD			childStartTime;
static	DWORD			childMinimumFinishTime;

static	int				childWidth;
static	int				childHeight;
static	int				childLine;
static	unsigned long	*childData;
static	Environment		*childMoveEnvironment;

static	void	childSetupBuild(int isFast);

extern	int				_myHeight;
extern	int				_width;
extern	unsigned long	*mapsData[2];
extern	int				mapEndTime;

//////////////////////////////////////////////////////////////////////
//
// Some basic communication functions
//
//////////////////////////////////////////////////////////////////////

static	void
threadChildCheckMessages(DWORD Milli)
{
	if (WaitForSingleObject(subEventParentToChild, Milli) != WAIT_OBJECT_0)
		return;

	switch (msgParentToChild)
	{
	case P2C_START_BUILD:
		childSetupBuild(0);
		break;
	case P2C_START_FAST_BUILD:
		childSetupBuild(1);
		break;
	case P2C_STOP_BUILD:
		childBuilding = 0;
		break;
	case P2C_EXIT_THREAD:
		childExiting = 1;
		break;
	}

	SetEvent(subEventACKParentToChild);
}
		

static	void
threadParentCheckMessages(DWORD Milli)
{
	if (WaitForSingleObject(subEventChildToParent, Milli) != WAIT_OBJECT_0)
		return;

	switch (msgChildToParent)
	{
	case C2P_FINISHED:
		parentBuilding = 0;
		break;
	}

	SetEvent(subEventACKChildToParent);
}

static	void		
threadSendMessageToChild(int msg)
{
	while (WaitForSingleObject(subMutex, 10) != WAIT_OBJECT_0)
		threadParentCheckMessages(0);

	msgParentToChild = msg;
	SetEvent(subEventParentToChild);
	WaitForSingleObject(subEventACKParentToChild, INFINITE);

	ReleaseMutex(subMutex);
}


static	int
threadSendMessageToParent(int msg)
{
	if (WaitForSingleObject(subMutex, 0) != WAIT_OBJECT_0)
		return -1;

	msgChildToParent = msg;
	SetEvent(subEventChildToParent);
	WaitForSingleObject(subEventACKChildToParent, INFINITE);

	ReleaseMutex(subMutex);

	return 0;
}


//////////////////////////////////////////////////////////////////////
//
// The parent functions
//
//////////////////////////////////////////////////////////////////////
void
parentStopCurrentBuild(void)
{
	threadSendMessageToChild(P2C_STOP_BUILD);
	parentBuilding = 0;
}


void
parentStartNewBuild(int fast)
{
	threadSendMessageToChild(fast ? P2C_START_FAST_BUILD: P2C_START_BUILD);
	parentBuilding = 1;
}


int
parentIsBuilding(void)
{
	threadParentCheckMessages(0);

	return parentBuilding;
}


void
parentShutdownThread(void)
{
	threadSendMessageToChild(P2C_EXIT_THREAD);
	WaitForSingleObject(subThread, INFINITE);
}

//////////////////////////////////////////////////////////////////////
//
// The child functions
//
//////////////////////////////////////////////////////////////////////


static	void
childSetupBuild(int isFast)
{
	// We need to grab all the information we need so we can go and do a build

	childBuilding = 1;

	childHeight = _myHeight;
	childWidth = _width;
	childData = mapsData[1];
	childStartTime = GetTickCount();
	if (isFast)
		childMinimumFinishTime = 1000;
	else
		childMinimumFinishTime = mapEndTime - childStartTime - 250;

	childLine = 0;

	childMoveEnvironment = ChooseMoveScript();
	
	if (childMoveEnvironment)
			SetupForRun(childMoveEnvironment);
}


static	void
childBuildNextLine(void)
{
	if (childLine >= childHeight)
	{
		if (threadSendMessageToParent(C2P_FINISHED) == 0)
			childBuilding = 0;

		return;
	}

	if (childMoveEnvironment)
			RunForLine(childMoveEnvironment, childData, childWidth, childHeight, childLine);

	childLine++;
	
	return;
}


static	void
childMainLoop(void *dummy)
{
	int	milli;
	int	goaltime;
	int	now;

	// Need to reseed the random number generator
	srand(GetTickCount());

	childExiting = 0;
	childBuilding = 0;

	while (1)
	{
		now = GetTickCount();

		if (childBuilding == 0)
			milli = INFINITE;
		else
		{
			goaltime = childStartTime + childMinimumFinishTime * childLine / childHeight;
			if (goaltime > now)
				milli = goaltime - now;
			else
				milli = 0;
		}

		threadChildCheckMessages(milli);

		if (childExiting)
			return;

		if (childBuilding)
			childBuildNextLine();
	}
}


void
threadSetupSubThread(void)
{
	SECURITY_ATTRIBUTES	sa;

	sa.nLength = sizeof(sa);
	sa.bInheritHandle = TRUE;
	
	subMutex = CreateMutex(&sa, FALSE, NULL);

	subEventChildToParent = CreateEvent(&sa, FALSE, FALSE, NULL);
	subEventACKChildToParent = CreateEvent(&sa, FALSE, FALSE, NULL);
	subEventParentToChild = CreateEvent(&sa, FALSE, FALSE, NULL);
	subEventACKParentToChild = CreateEvent(&sa, FALSE, FALSE, NULL);

	// Setup mutexes and conditional things
	parentBuilding = 0;

	subThread = (HANDLE)_beginthread(childMainLoop, 0, NULL);
	// Not 100% sure about the priority of the sub thread, but I think it should play second fiddle
	// hmm.. changed my mind..
	// SetThreadPriority(subThread, THREAD_PRIORITY_BELOW_NORMAL);
}
#endif
