//////////////////////////////////////////////////////////////////////////////////////
//
// font stuff
//
// I got all excited and did a different font system.  I don't think it's
// actually very good, nor do I think it's particularly efficient, but I'm
// hoping it will look pretty nice.  Let's try it, shall we.
//
//////////////////////////////////////////////////////////////////////////////////////


extern	unsigned long	fontBitmap[];
extern	int		fontOffsets[3][33];

// Font Bitmap is 2048 * 258

__inline int
fontGetPoint(unsigned int x, unsigned int y)
{
	int byte = y * 64 + (x >> 5);
	unsigned long mask = (unsigned long)0x80000000 >> (x & 0x1f);

	return ((*(fontBitmap + byte)) & mask);
}

__inline int
fontGetAAPoint(unsigned int x, unsigned int y, int xStart, int yStart)
{
	int v1 = fontGetPoint(xStart + (x >> 16), yStart + (y >> 16)) ? 255 : 0;
	int v2 = fontGetPoint(xStart + (x >> 16) + 1, yStart + (y >> 16)) ? 255 : 0;
	int v3 = fontGetPoint(xStart + (x >> 16), yStart + (y >> 16) + 1) ? 255 : 0;
	int v4 = fontGetPoint(xStart + (x >> 16) + 1, yStart + (y >> 16) + 1) ? 255 : 0;

	v1 = v1 + (((v2 - v1) * (x & 0xffff)) >> 16);
	v3 = v3 + (((v4 - v3) * (x & 0xffff)) >> 16);

	return v1 + (((v3 - v1) * (y & 0xffff)) >> 16);
}


int
fontDraw(int xBase, int yBase, int height, unsigned char c, int color)
{
	int	line = (c >> 5) - 1;
	int	let = c & 0x1f;
	int	width;
	int	xStart;
	int	yStart;
	int	y, x;
	unsigned int ry;
	int	tc;

	if (c < 32 || c > 127)
		return xBase;

	width = fontOffsets[line][let + 1] - fontOffsets[line][let];

	width = width * height / 86;

	yStart = line * 86;
	xStart = fontOffsets[line][let];

	for (y = 0; y < height; y++)
	{
		ry = y * 86 / height;
		for (x = 0; x < width; x++)
		{
			tc = fontGetPoint(x * 86 / height + xStart, ry + yStart);

			if (tc)
				effectsDrawPoint(x + xBase, y + yBase, color, 1);
		}
	}

	return xBase + width;
}

int
fontDrawDirect(int xBase, int yBase, int height, unsigned char c, unsigned long color)
{
	int	line = (c >> 5) - 1;
	int	let = c & 0x1f;
	int	width;
	int	xStart;
	int	yStart;
	int	y, x;
	unsigned int ry;
	int	tc;

	if (c < 32 || c > 127)
		return xBase;

	width = fontOffsets[line][let + 1] - fontOffsets[line][let];

	width = width * height / 86;

	yStart = line * 86;
	xStart = fontOffsets[line][let];

	for (y = 0; y < height; y++)
	{
		ry = y * 86 / height;
		for (x = 0; x < width; x++)
		{
			tc = fontGetPoint(x * 86 / height + xStart, ry + yStart);

			if (tc)
				//effectsDrawPoint(x + xBase, y + yBase, color, 1);
				effectsDrawDirectPoint(x + xBase, y + yBase, color);
		}
	}

	return xBase + width;
}

int
fontDrawAA(int xBase, int yBase, int height, unsigned char c, int color)
{
	int	line = (c >> 5) - 1;
	int	let = c & 0x1f;
	int	width;
	int	xStart;
	int	yStart;
	int	y, x;
	unsigned int ry;
	int	tc;

	if (c < 32 || c > 127)
		return xBase;

	width = fontOffsets[line][let + 1] - fontOffsets[line][let];

	width = width * height / 86;

	yStart = line * 86;
	xStart = fontOffsets[line][let];

	for (y = 0; y < height; y++)
	{
		ry = (y << 16) * 86 / height;
		for (x = 0; x < width; x++)
		{
			tc = fontGetAAPoint((x << 16) * 86 / height, ry, xStart, yStart);

			if (tc)
				effectsDrawPoint(x + xBase, y + yBase, tc * color / 255, 1);
		}
	}

	return xBase + width;
}


__inline int
fontCharacterWidth(int height, char c)
{
	int	line = (c >> 5) - 1;
	int	let = c & 0x1f;

	return (fontOffsets[line][let + 1] - fontOffsets[line][let]) * height / 86;
}

int
fontStringLength(int height, char *text)
{
	int	len = 0;

	while (*text)
		len += fontCharacterWidth(height, *(text++));

	return len;
}


int
fontDrawDirectAA(int xBase, int yBase, int height, unsigned char c, unsigned long color)
{
	int	line = (c >> 5) - 1;
	int	let = c & 0x1f;
	int	width;
	int	xStart;
	int	yStart;
	int	y, x;
	unsigned int ry;
	int	tc;

	if (c < 32 || c > 127)
		return xBase;

	width = fontOffsets[line][let + 1] - fontOffsets[line][let];

	width = width * height / 86;

	yStart = line * 86;
	xStart = fontOffsets[line][let];

	for (y = 0; y < height; y++)
	{
		ry = (y << 16) * 86 / height;
		for (x = 0; x < width; x++)
		{
			tc = fontGetAAPoint((x << 16) * 86 / height, ry, xStart, yStart);

			if (tc)
				//effectsDrawDirectPoint(x + xBase, y + yBase,  tc * r / 255, tc * g / 255, tc * b / 255);
				effectsDrawDirectPoint(x + xBase, y + yBase, color); //  <-- doesn't do any AA currently
		}
	}

	return xBase + width;
}
//effectsDrawDirectPoint(int x, int y, int r, int g, int b)

int
fontDrawText(int xBase, int yBase, int height, char *text, int color)
{
	while (*text)
		xBase = fontDraw(xBase, yBase, height, *(text++), color);

	return xBase;
}

int
fontDrawTextDirect(int xBase, int yBase, int height, char *text, unsigned long color)
{
	while (*text)
		xBase = fontDrawDirect(xBase, yBase, height, *(text++), color);

	return xBase;
}

int
fontDrawTextAA(int xBase, int yBase, int height, char *text, int color)
{
	while (*text)
		xBase = fontDrawAA(xBase, yBase, height, *(text++), color);

	return xBase;
}

int
fontDrawTextDirectAA(int xBase, int yBase, int height, char *text, unsigned long color)
{
	while (*text)
		xBase = fontDrawDirectAA(xBase, yBase, height, *(text++), color);

	return xBase;
}
//fontDrawDirectAA(int xBase, int yBase, int height, unsigned char c, int r, int g, int b)
