#include "smear.h"

/* Example to get a decent FONT

Using a Stock Font to Draw Text
The system provides six stock fonts. A stock font is a logical font that an application can obtain by calling the GetStockObject function and specifying the requested font. The following list contains the values that you can specify to obtain a stock font. 

Value Meaning 
ANSI_FIXED_FONT Specifies a monospace font based on the Windows character set. A Courier font is typically used. 
ANSI_VAR_FONT Specifies a proportional font based on the Microsoft Windows character set. MS Sans Serif is typically used. 
DEVICE_DEFAULT_FONT Specifies the preferred font for the specified device. This is typically the System font for display devices; however, for some dot-matrix printers this is a font that is resident on the device. (Printing with this font is usually faster than printing with a downloaded, bitmap font). 
OEM_FIXED_FONT Specifies a monospace font based on an OEM character set. For IBM computers and compatibles, the OEM font is based on the IBM PC character set. 
SYSTEM_FONT Specifies the System font. This is a proportional font based on the Windows character set, and is used by the operating system to display window titles, menu names, and text in dialog boxes. The System font is always available. Other fonts are available only if they have been installed. 
SYSTEM_FIXED_FONT Specifies a monospace font compatible with the System font in Windows versions earlier than 3.0. 

The following example retrieves a handle to the variable stock font, selects it into a device context, and then writes a string using that font: 

HFONT hfnt, hOldFont; 
 
    hfnt = GetStockObject(ANSI_VAR_FONT); 
    if (hOldFont = SelectObject(hdc, hfnt)) 
    { 
        TextOut(hdc, 10, 50, "Sample ANSI_VAR_FONT text.", 26); 
        SelectObject(hdc, hOldFont); 
    } 

---------------------------------------
TextOut
The TextOut function writes a character string at the specified location, using the currently selected font, background color, and text color. 

BOOL TextOut(
  HDC hdc,           // handle to DC
  int nXStart,       // x-coordinate of starting position
  int nYStart,       // y-coordinate of starting position
  LPCTSTR lpString,  // character string
  int cbString       // number of characters
);
Parameters
hdc 
[in] Handle to the device context. 
nXStart 
[in] Specifies the logical x-coordinate of the reference point that the system uses to align the string. 
nYStart 
[in] Specifies the logical y-coordinate of the reference point that the system uses to align the string. 
lpString 
[in] Pointer to the string to be drawn. The string does not need to be zero-terminated, since cbString specifies the length of the string. 
cbString 
[in] Specifies the number of characters in the string. 
Windows 95/98: This value may not exceed 8192. 
If the function fails, the return value is zero. 

Example:
// Obtain the window's client rectangle 
GetClientRect(hwnd, &r);  
// THE FIX: by setting the background mode 
// to transparent, the region is the text itself 
// SetBkMode(hdc, TRANSPARENT);  
// Bracket begin a path 
BeginPath(hdc);  
// Send some text out into the world 
TextOut(hdc, r.left, r.top, "Defenestration can be hazardous", 4);  
// Bracket end a path 
EndPath(hdc);  
// Derive a region from that path 
SelectClipPath(hdc, RGN_AND);  
// This generates the same result as SelectClipPath() 
// SelectClipRgn(hdc, PathToRegion(hdc));  
// Fill the region with grayness 
FillRect(hdc, &r, GetStockObject(GRAY_BRUSH)); 
*/

/*
-----------------------------
COLORREF SetTextColor(
  HDC hdc,           // handle to DC
  COLORREF crColor   // text color
);
When specifying an explicit RGB color, the COLORREF value has the following hexadecimal form: 
0x00bbggrr 
(use RGB:
COLORREF RGB(
  BYTE byRed,    // red component of color
  BYTE byGreen,  // green component of color
  BYTE byBlue    // blue component of color
);
)
----------------------------
SetBkColor
The SetBkColor function sets the current background color to the specified color value, or to the nearest physical color if the device cannot represent the specified color value. 

COLORREF SetBkColor(
  HDC hdc,           // handle to DC
  COLORREF crColor   // background color value
);
Parameters
hdc 
[in] Handle to the device context. 
crColor 
[in] Specifies the new background color. To make a COLORREF value, use the RGB macro. 
Return Values
If the function succeeds, the return value specifies the previous background color as a COLORREF value. 
If the function fails, the return value is CLR_INVALID. 
----------------------------
SetTextAlign
The SetTextAlign function sets the text-alignment flags for the specified device context. 

UINT SetTextAlign(
  HDC hdc,     // handle to DC
  UINT fMode   // text-alignment option
);
Parameters
hdc 
[in] Handle to the device context. 
fMode 
[in] Specifies the text alignment by using a mask of the values in the following list. Only one flag can be chosen from those that affect horizontal and vertical alignment. In addition, only one of the two flags that alter the current position can be chosen. Value Meaning 
TA_BASELINE The reference point will be on the base line of the text. 
TA_BOTTOM The reference point will be on the bottom edge of the bounding rectangle. 
TA_TOP The reference point will be on the top edge of the bounding rectangle. 
TA_CENTER The reference point will be aligned horizontally with the center of the bounding rectangle. 
TA_LEFT The reference point will be on the left edge of the bounding rectangle. 
TA_RIGHT The reference point will be on the right edge of the bounding rectangle. 
TA_NOUPDATECP The current position is not updated after each text output call. The reference point is passed to the text output function. 
TA_RTLREADING Middle-Eastern Windows: The text is laid out in right to left reading order, as opposed to the default left to right order. This applies only when the font selected into the device context is either Hebrew or Arabic. 
TA_UPDATECP The current position is updated after each text output call. The current position is used as the reference point. 
When the current font has a vertical default base line, as with Kanji, the following values must be used instead of TA_BASELINE and TA_CENTER. Value Meaning 
VTA_BASELINE The reference point will be on the base line of the text. 
VTA_CENTER The reference point will be aligned vertically with the center of the bounding rectangle. 
The default values are TA_LEFT, TA_TOP, and TA_NOUPDATECP. 
Return Values
If the function succeeds, the return value is the previous text-alignment setting.
If the function fails, the return value is GDI_ERROR.
SetTextAlign (hdc, GetTextAlign(hdc) & (~TA_CENTER))
or 
SetTextAlign (hdc,TA_LEFT | <other flags>)
You can also use SetTextAlign (hdc, TA_LEFT) for this purpose, but this loses any vertical or right-to-left settings.
----------------------------------------
GetPixel
The GetPixel function retrieves the red, green, blue (RGB) color value of the pixel at the specified coordinates. 

COLORREF GetPixel(
  HDC hdc,    // handle to DC
  int nXPos,  // x-coordinate of pixel
  int nYPos   // y-coordinate of pixel
);
Parameters
hdc 
[in] Handle to the device context. 
nXPos 
[in] Specifies the logical x-coordinate of the pixel to be examined. 
nYPos 
[in] Specifies the logical y-coordinate of the pixel to be examined. 
Return Values
The return value is the RGB value of the pixel. If the pixel is outside of the current clipping region, the return value is CLR_INVALID. 

*/

#define ADDALPHA(a, col) (col | ((unsigned long)a<<24))

#define	MAX_FONT_LAYERS	3

HDC		wfDC[MAX_FONT_LAYERS];
BITMAPINFO  	bitmapinfo[MAX_FONT_LAYERS];
HBITMAP		wftxtbuf[MAX_FONT_LAYERS];
HBITMAP		wftxtbufold[MAX_FONT_LAYERS];
void*		wfmembuf[MAX_FONT_LAYERS];
HFONT		hfont[MAX_FONT_LAYERS] = { NULL, NULL, NULL };
HFONT		hOldFont[MAX_FONT_LAYERS];

int		wsWidth[MAX_FONT_LAYERS] = { 0, 0, 0 };
int		wsHeight[MAX_FONT_LAYERS] = { 0, 0, 0 };
int		wsDeinit[MAX_FONT_LAYERS] = { 0, 0, 0 };

unsigned char	*_textHold[MAX_FONT_LAYERS] =
	{ (unsigned char *)0, (unsigned char *)0, (unsigned char *)0 };

int
WSInit(int width, int height, int layer)
{
	if (width != wsWidth[layer] || height != wsHeight[layer])
	{
		if (_textHold[layer])
			free(_textHold[layer]);

		_textHold[layer] = (unsigned char *)malloc(width * height * sizeof(unsigned char));

		wsDeinit[layer] = 1;
	}

	if (wsDeinit[layer])
	{
		if (wsWidth[layer] != 0 || wsHeight[layer] != 0)
		{
			DeleteObject(wftxtbuf[layer]);
			DeleteDC(wfDC[layer]);
		}

		wsWidth[layer] = width;
		wsHeight[layer] = height;

		wfDC[layer] = CreateCompatibleDC(NULL);

		bitmapinfo[layer].bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		bitmapinfo[layer].bmiHeader.biWidth = width;
		bitmapinfo[layer].bmiHeader.biHeight = height; /* top-down */
		bitmapinfo[layer].bmiHeader.biPlanes = 1;
		bitmapinfo[layer].bmiHeader.biBitCount = 32;
		bitmapinfo[layer].bmiHeader.biCompression = BI_RGB;
		bitmapinfo[layer].bmiHeader.biSizeImage = 0;
		bitmapinfo[layer].bmiHeader.biClrUsed = 256;
		bitmapinfo[layer].bmiHeader.biClrImportant = 256;
		wftxtbuf[layer] = CreateDIBSection(wfDC[layer], &(bitmapinfo[layer]), DIB_RGB_COLORS, &(wfmembuf[layer]), 0, 0);
		wftxtbufold[layer] = SelectObject(wfDC[layer], wftxtbuf[layer]);

		SetBkColor(wfDC[layer], 0x00000000);   // background color value
		SetBkMode(wfDC[layer], TRANSPARENT);
		SetTextColor(wfDC[layer], 0x00999999);
		ExtFloodFill(wfDC[layer],(width/2),(height/2),0x00000000,FLOODFILLBORDER);
		wsDeinit[layer] = 0;

		return 1;
	}

	return 0;
}


void
WSDeInit(int layer)
{
	wsDeinit[layer] = 1;
}

void
WSSetFont(COLORREF textcolor, char* fontface, int pointsize, int weight, int layer)
{
    CHOOSEFONT cf; 
    LOGFONT lf; 
 
    // Initialize members of the CHOOSEFONT structure. 
 
    cf.lStructSize = sizeof(CHOOSEFONT); 
    cf.hwndOwner = (HWND)NULL; 
    cf.hDC = (HDC)NULL; 
    cf.lpLogFont = &lf; 
    cf.iPointSize = pointsize; 
    cf.Flags = CF_SCREENFONTS; 
    cf.rgbColors = textcolor; 
    cf.lCustData = 0L; 
    cf.lpfnHook = (LPCFHOOKPROC)NULL; 
    cf.lpTemplateName = (LPSTR)NULL; 
    cf.hInstance = (HINSTANCE) NULL; 
    cf.lpszStyle = (LPSTR)NULL; 
    cf.nFontType = SCREEN_FONTTYPE; 
    cf.nSizeMin = 0; 
    cf.nSizeMax = 0; 

    cf.lpLogFont->lfHeight = -MulDiv(cf.iPointSize, GetDeviceCaps(wfDC[layer], LOGPIXELSY), 72);
    cf.lpLogFont->lfWidth = 0;
    cf.lpLogFont->lfEscapement = 0;
    cf.lpLogFont->lfOrientation = 0;
    cf.lpLogFont->lfWeight = weight;
    cf.lpLogFont->lfItalic = FALSE;
    cf.lpLogFont->lfUnderline = FALSE;
    cf.lpLogFont->lfStrikeOut = FALSE;
    cf.lpLogFont->lfCharSet = ANSI_CHARSET;
    cf.lpLogFont->lfOutPrecision = OUT_DEFAULT_PRECIS;
    cf.lpLogFont->lfClipPrecision = CLIP_DEFAULT_PRECIS;
    cf.lpLogFont->lfQuality = PROOF_QUALITY|NONANTIALIASED_QUALITY;
    cf.lpLogFont->lfPitchAndFamily = FF_SWISS;
    strcpy(cf.lpLogFont->lfFaceName, fontface);

    //Choose font here

    if (hfont[layer])
	    DeleteObject(hfont[layer]);

    hfont[layer] = CreateFontIndirect(cf.lpLogFont); 

    hOldFont[layer] = SelectObject(wfDC[layer], hfont[layer]);
}


int
WSSupDrawText(int cX, int cY, const char* satz, int layer, int align)
{
	char	tmp_string[1024];
	char	*p;

	strcpy(tmp_string, satz);

	for (p = tmp_string; *p; p++)
		if (*p == '_')
			*p = ' ';

	SetTextColor(wfDC[layer], 0x00ffffff);
	SetTextAlign(wfDC[layer], align);
	return (TextOut(wfDC[layer], cX, cY, tmp_string, strlen(tmp_string)) != 0);
}


void
WSCopyOut(int layer)
{
	unsigned char	*l;
	int				x, y;

	for (l = _textHold[layer], y = 0; y < wsHeight[layer]; y++)
		for (x = 0; x < wsWidth[layer]; x++, l++)
			*l = (unsigned char)(GetPixel(wfDC[layer], x, y) & 0xff);
}


void
WSCopyHoldToVid(unsigned long *Video, int pitch, int offset, unsigned long color, int layer)
{
	unsigned char	*ip;
	unsigned long	*op;
	int				x, y;

	Video += offset * pitch;

	for (ip = _textHold[layer], y = 0; y < wsHeight[layer]; y++, Video += pitch)
		for (op = Video, x = 0; x < wsWidth[layer]; x++, op++, ip++)
			if (*ip)
				*op = color;
}


void
WSCopyHoldToSB(unsigned char *SB, int color, int layer)
{
	unsigned char	*ip;
	unsigned char	*op;
	int				x, y;

	for (ip = _textHold[layer], op = SB, y = 0; y < wsHeight[layer]; y++)
		for (x = 0; x < wsWidth[layer]; x++, op++, ip++)
			if (*ip)
				*op = color;
}
